using System;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Media3D;

namespace TestProject
{
    /// <summary>
    /// ü ޽
    /// </summary>
    public class SphereMesh : MeshGeneratorBase
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        /// ߽ Ӽ
        /// </summary>
        public static readonly DependencyProperty CenterPointProperty = DependencyProperty.Register
        (
            "CenterPoint",
            typeof(Point3D),
            typeof(SphereMesh),
            new PropertyMetadata(new Point3D(), PropertyChanged)
        );

        /// <summary>
        /// ݰ Ӽ
        /// </summary>
        public static readonly DependencyProperty RadiusProperty = DependencyProperty.Register
        (
            "Radius",
            typeof(double),
            typeof(SphereMesh),
            new PropertyMetadata(1.0, PropertyChanged)
        );

        /// <summary>
        /// ̽  Ӽ
        /// </summary>
        public static readonly DependencyProperty SliceCountProperty = DependencyProperty.Register
        (
            "SliceCount",
            typeof(int),
            typeof(SphereMesh),
            new PropertyMetadata(36, PropertyChanged),
            ValidateSliceCount
        );

        /// <summary>
        ///   Ӽ
        /// </summary>
        public static readonly DependencyProperty StackCountProperty = DependencyProperty.Register
        (
            "StackCount",
            typeof(int),
            typeof(SphereMesh),
            new PropertyMetadata(18, PropertyChanged),
            ValidateStackCount
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region ߽ - CenterPoint

        /// <summary>
        /// ߽
        /// </summary>
        public Point3D CenterPoint
        {
            set
            {
                SetValue(CenterPointProperty, value);
            }
            get
            {
                return (Point3D)GetValue(CenterPointProperty);
            }
        }

        #endregion
        #region ݰ - Radius

        /// <summary>
        /// ݰ
        /// </summary>
        public double Radius
        {
            set
            {
                SetValue(RadiusProperty, value);
            }
            get
            {
                return (double)GetValue(RadiusProperty);
            }
        }

        #endregion
        #region ̽  - SliceCount

        /// <summary>
        /// ̽ 
        /// </summary>
        public int SliceCount
        {
            set
            {
                SetValue(SliceCountProperty, value);
            }
            get
            {
                return (int)GetValue(SliceCountProperty);
            }
        }

        #endregion
        #region   - StackCount

        /// <summary>
        ///  
        /// </summary>
        public int StackCount
        {
            set
            {
                SetValue(StackCountProperty, value);
            }
            get
            {
                return (int)GetValue(StackCountProperty);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region  - SphereMesh()

        /// <summary>
        /// 
        /// </summary>
        public SphereMesh()
        {
            PropertyChanged(new DependencyPropertyChangedEventArgs());
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region ̽  Ἲ üũϱ - ValidateSliceCount(sourceObject)

        /// <summary>
        /// ̽  Ἲ üũϱ
        /// </summary>
        /// <param name="sourceObject">ҽ ü</param>
        /// <returns>ó </returns>
        private static bool ValidateSliceCount(object sourceObject)
        {
            return (int)sourceObject > 2;
        }

        #endregion
        #region   Ἲ üũϱ - ValidateStackCount(sourceObject)

        /// <summary>
        ///   Ἲ üũϱ
        /// </summary>
        /// <param name="sourceObject">ҽ ü</param>
        /// <returns>ó </returns>
        private static bool ValidateStackCount(object sourceObject)
        {
            return (int)sourceObject > 1;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Protected

        #region ﰢ ϱ - Triangulate(e, vertexCollection, normalCollection, indexCollection, textureCollection)

        /// <summary>
        /// ﰢ ϱ
        /// </summary>
        /// <param name="e">̺Ʈ </param>
        /// <param name="vertexCollection"> ÷</param>
        /// <param name="normalCollection"> ÷</param>
        /// <param name="indexCollection">ε ÷</param>
        /// <param name="textureCollection">ؽó ÷</param>
        protected override void Triangulate(DependencyPropertyChangedEventArgs e, Point3DCollection vertexCollection, Vector3DCollection normalCollection, Int32Collection indexCollection, PointCollection textureCollection)
        {
            vertexCollection.Clear();
            normalCollection.Clear();
            indexCollection.Clear();
            textureCollection.Clear();

            for(int stack = 0; stack <= StackCount; stack++)
            {
                double phi   = Math.PI / 2 - stack * Math.PI / StackCount;
                double y     = Radius * Math.Sin(phi);
                double scale = -Radius * Math.Cos(phi);

                for(int slice = 0; slice <= SliceCount; slice++)
                {
                    double theta = slice * 2 * Math.PI / SliceCount;
                    double x     = scale * Math.Sin(theta);
                    double z     = scale * Math.Cos(theta);

                    Vector3D normal = new Vector3D(x, y, z);

                    normalCollection.Add(normal);

                    vertexCollection.Add(normal + CenterPoint);

                    textureCollection.Add(new Point((double)slice / SliceCount, (double)stack / StackCount));
                }
            }

            for(int stack = 0; stack < StackCount; stack++)
            {
                for(int slice = 0; slice < SliceCount; slice++)
                {
                    if(stack != 0)
                    {
                        indexCollection.Add((stack + 0) * (SliceCount + 1) + slice    );
                        indexCollection.Add((stack + 1) * (SliceCount + 1) + slice    );
                        indexCollection.Add((stack + 0) * (SliceCount + 1) + slice + 1);
                    }

                    if(stack != StackCount - 1)
                    {
                        indexCollection.Add((stack + 0) * (SliceCount + 1) + slice + 1);
                        indexCollection.Add((stack + 1) * (SliceCount + 1) + slice    );
                        indexCollection.Add((stack + 1) * (SliceCount + 1) + slice + 1);
                    }
                }
            }
        }

        #endregion
        #region νϽ ϱ (ھ) - CreateInstanceCore()

        /// <summary>
        /// νϽ ϱ (ھ)
        /// </summary>
        /// <returns>νϽ</returns>
        protected override Freezable CreateInstanceCore()
        {
            return new SphereMesh();
        }

        #endregion
    }
}